import unittest
import shutil
from pathlib import Path


class TestCCMetagen(unittest.TestCase):
    def test_import_ccmetagen_module(self):
        try:
            import ccmetagen
            _ = f"CCMetagen version {ccmetagen.__version__}." # noqa
        except ImportError:
            self.fail("Failed to import ccmetagen module")

    def test_ccmetagen_scripts_on_path(self):
        scripts = ["CCMetagen.py", "CCMetagen_extract_seqs.py", "CCMetagen_merge.py"]
        for script_name in scripts:
            try:
                # Check if the script is on the PATH
                script_path = shutil.which(script_name)
                self.assertIsNotNone(script_path, f"{script_name} not found on PATH")

                # Check if the script can be executed
                script_path = Path(script_path)
                self.assertTrue(script_path.is_file(), f"{script_name} is not a file")
                self.assertTrue(script_path.exists(), f"{script_name} does not exist")

            except Exception as e:
                self.fail(f"Unexpected error: {e}")


if __name__ == "__main__":
    unittest.main()



__version__ = "1.5.0"

# This can be configured
_RANK_THRESHOLDS = {
    "species_threshold": 98.41, # Yeast - Vu et al 2016
    "genus_threshold": 96.31,   # Yeast - Vu et al 2016
    "family_threshold": 88.51,  # Filamentous fungi - Vu et al 2019
    "order_threshold": 81.21,   # Filamentous fungi - Vu et al 2019
    "class_threshold": 80.91,   # Filamentous fungi - Vu et al 2019
    "phylum_threshold": 0,      # no data, no filtering    
}
_ARGPARSE_DEFAULTS = {
    "mode": "both",
    "output_fp": "CCMetagen_out",
    "reference_database": "nt",
    "extended_output_file": "n",
    "depth_unit": "kma",
    "mapstat": None,
    "depth": 0.2,
    "coverage": 20,
    "query_identity": 50,
    "pvalue": 0.05,
    "krona_mode": "Depth",
    "local_taxfile": None,
    "turn_off_sim_thresholds": "n",
    **_RANK_THRESHOLDS,
}

__all__ = ["__version__", "_ARGPARSE_DEFAULTS", "_RANK_THRESHOLDS"]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Classes that store taxonomic info from matches and NCBI lineage

@ V.R.Marcelino
Created on 12 Jul 2018.
"""


### Stores taxid and other info
class TaxInfo:
    def __init__(
        self,
        TaxId=None,
        Lineage=None,
        Sample=None,
        RefDatabase=None,
        Abundance=None,
        Superkingdom=None,
        Superkingdom_TaxId=None,
        Kingdom=None,
        Kingdom_TaxId=None,
        Phylum=None,
        Phylum_TaxId=None,
        Class=None,
        Class_TaxId=None,
        Order=None,
        Order_TaxId=None,
        Family=None,
        Family_TaxId=None,
        Genus=None,
        Genus_TaxId=None,
        Species=None,
        Species_TaxId=None,
        LCA_TaxId=None,
        Coverage=None,
    ):
        # info from matches
        self.TaxId = TaxId
        self.Lineage = Lineage  # the results of the match, not considering taxIDs
        self.Sample = Sample
        self.RefDatabase = RefDatabase
        self.Abundance = Abundance

        # info from NCBI
        self.Superkingdom = Superkingdom
        self.Superkingdom_TaxId = Superkingdom_TaxId
        self.Kingdom = Kingdom
        self.Kingdom_TaxId = Kingdom_TaxId
        self.Phylum = Phylum
        self.Phylum_TaxId = Phylum_TaxId
        self.Class = Class
        self.Class_TaxId = Class_TaxId
        self.Order = Order
        self.Order_TaxId = Order_TaxId
        self.Family = Family
        self.Family_TaxId = Family_TaxId
        self.Genus = Genus
        self.Genus_TaxId = Genus_TaxId
        self.Species = Species
        self.Species_TaxId = Species_TaxId

        # Taxid of the Lowest Common Ancestor
        self.LCA_TaxId = LCA_TaxId

        # info from Fungi Table only
        self.Coverage = Coverage



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
CCMetagen_merge.py

Merge the abundance results of CCMetagen for several samples into a single table.
The results must be in .csv format (default CCMetagen or ---mode text).
The files must end in ".ccm.csv".


USAGE example 1: Merge table at family level:
CCMetagen_merge.py -i CCMetagen_folder -t Family

USAGE example 2: Merge table by species (default) and keep only Cryptococcus and Candida:
CCMetagen_merge.py -i CCMetagen_folder -kr k -l Genus -tlist Candida,Cryptococcus

@ V.R.Marcelino
Created on Tue Dec 18 10:00:48 2018.
"""

import sys
import pandas as pd
from argparse import ArgumentParser
import os


parser = ArgumentParser()
parser.add_argument(
    "-i",
    "--input_fp",
    help="""Path to the folder containing CCMetagen text results.
                    Note that files must end with ".csv"
                    and the folder should not contain other .csv files""",
    required=True,
)
parser.add_argument(
    "-t",
    "--tax_level",
    default="Species",
    help="""Taxonomic level to merge the results. Options:
                    Closest_match (includes different genes for the same species),
                    Species (Default), Genus, Family, Order, Class, Phylum, Kingdom and Superkingdom
                    """,
    required=False,
)
parser.add_argument(
    "-o",
    "--output_fp",
    default="merged_samples",
    help="Path to the output file. Default = merged_samples",
    required=False,
)

# arguments to filter by taxonomy:
parser.add_argument(
    "-kr",
    "--keep_or_remove",
    default="n",
    help="keep or remove taxa. Options = k (keep), r (remove) and n (none, default)",
    required=False,
)

parser.add_argument(
    "-l",
    "--filtering_tax_level",
    default="none",
    help="level to perform taxonomic filtering, default = none",
    required=False,
)

parser.add_argument(
    "-tlist",
    "--taxa_list",
    default=[],
    type=str,
    help="list taxon names (comma-separated) that you want to keep or exclude",
    required=False,
)

args = parser.parse_args()
in_folder = args.input_fp
tax_level = args.tax_level
output = args.output_fp
kr = args.keep_or_remove


# debugging:
# in_folder = "/Users/vmar0011/Documents/Programs_dev/03_CCM_nt"
# tax_level = "Species"
# output = "merged_samples_depth_species_no_ecoli"
# kr = "r" # k for keep, r for remove, and n for none
# level = "Species"
# taxa = ["Escherichia coli"]


# Set the taxonomic levels to retain info:
l = [
    "Superkingdom",
    "Kingdom",
    "Phylum",
    "Class",
    "Order",
    "Family",
    "Genus",
    "Species",
    "Closest_match",
]

f4agg = {}

f4agg["Depth"] = "sum"
for i in l:
    if i == tax_level:
        f4agg[i] = "first"
        break
    else:
        f4agg[i] = "first"

# list of wanted taxa (no depth):
wanted_taxa = list(f4agg.keys())
wanted_taxa.remove("Depth")

# create new dataframe:
all_samples = pd.DataFrame()


# read input files and merge results
for file in os.listdir(in_folder):
    if file.endswith(".ccm.csv"):
        sample_name = file.split(".res.ccm.csv")[0]
        result_fp = os.path.join(in_folder, file)

        df = pd.read_csv(result_fp, sep=",", index_col=0)

        # this is needed because groupby excludes rows with NAs
        df = df.fillna("NA")

        # keep or remove taxa:
        if kr == "k":
            level = args.filtering_tax_level
            taxa = args.taxa_list.split(",")
            filt_df_str = "df[df['{}'].isin({})]".format(level, taxa)
            df = eval(filt_df_str)

        elif kr == "r":
            level = args.filtering_tax_level
            taxa = args.taxa_list.split(",")
            filt_df_str = "df[~df['{}'].isin({})]".format(level, taxa)
            df = eval(filt_df_str)
        elif kr == "n":
            pass
        else:
            print("kr must be k (keep), r (remove) or n (none).")
            sys.exit("Try again.")

        # if tax_level = closest_match, we need an extra column:

        if tax_level == "Closest_match":
            df["Closest_match"] = df.index
            df.index.name = None  # This is needed in modern versions of Pandas (1.0.1)

        depth_by_tax = df.groupby(by=wanted_taxa).agg(f4agg)
        depth_by_tax.rename(columns={"Depth": sample_name}, inplace=True)

        all_samples = pd.concat([all_samples, depth_by_tax], sort=True, axis=1)


# Group taxon ranks
all_samples = all_samples.groupby(by=all_samples.columns, axis=1).first()

# add taxon info at the end of the table:
tax_cols_l = list(f4agg.keys())
tax_cols_l.remove("Depth")

tax_cols = all_samples[tax_cols_l]
sample_cols = all_samples.drop(columns=tax_cols_l)

all_samples = pd.merge(sample_cols, tax_cols, left_index=True, right_index=True)

# Fill NaN with zeros:
all_samples = all_samples.fillna(0)

# Remove the artificially added NAs:
all_samples = all_samples.replace(["NA"], value="")


### Save
out = output + ".csv"
pd.DataFrame.to_csv(all_samples, out, index=False)

print("Done. Abundance estimates for all samples saved as %s" % (out))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Function that takes a taxid as input and returns a class containing the taxonomic info and taxids
for multiple taxonomic ranks.

Required to parse the results of KMA and store them in the SQLite3 'bench.db'

@ V.R.Marcelino
Created on 12 Jul 2018.
"""

from ete3 import NCBITaxa


def lineage_extractor(query_taxid, TaxInfo_object, taxfile=None):
    list_of_taxa_ranks = [
        "Superkingdom",
        "Kingdom",
        "Phylum",
        "Class",
        "Order",
        "Family",
        "Genus",
        "Species",
    ]
    if taxfile is not None:
        ncbi = NCBITaxa(taxfile)
    else:
        ncbi = NCBITaxa()
    lineage = ncbi.get_lineage(query_taxid)
    ranks = ncbi.get_rank(lineage)
    names = ncbi.get_taxid_translator(lineage)

    # get known data
    for key, val in ranks.items():
        for rank in list_of_taxa_ranks:
            if val.lower() == rank.lower():
                setattr(TaxInfo_object, rank, names[key])
                setattr(TaxInfo_object, rank + "_TaxId", key)

    # fill in the blanks
    for attr in list_of_taxa_ranks:
        if getattr(TaxInfo_object, attr) is None:
            initial = attr.lower()[0] if attr != "Superkingdom" else "sk"
            setattr(TaxInfo_object, attr, "unk_" + initial)

    return TaxInfo_object



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Script to extract sequence reads (or contigs) assigned to a given taxon.

Created on Wed Jun 10 19:37:36 2020.

@author: V.R.Marcelino
"""


# imports

import sys
import pandas as pd
import csv
from argparse import ArgumentParser


# help
if len(sys.argv) == 1:
    print("")
    print(
        """Produces a fasta file containing the sequences assigned to a given taxon"""
    )
    print("For help and options, type: CCMetagen_extract_seqs.py -h")
    print("")


parser = ArgumentParser()
parser.add_argument(
    "-ifrag",
    "--input_frags",
    help="The path to the file containing frags (decompressed)",
    required=True,
)
parser.add_argument(
    "-iccm",
    "--input_ccmetagen",
    help="The path to the ccmetagen result csv file",
    required=True,
)
parser.add_argument(
    "-l",
    "--taxonomic_level",
    help="""Taxonomic level to merge the results. Options:
                    Species, Genus, Family, Order, Class, Phylum, Kingdom and Superkingdom""",
    required=True,
)
parser.add_argument(
    "-t",
    "--taxon",
    help="""Taxon for which you want to extract sequences. 
                    Use quotation marks to specify species (e.g. -t 'Escherichia coli')""",
    required=True,
)
parser.add_argument(
    "-o",
    "--output_fp",
    default="wanted_taxon_seqs",
    help="Path to the output file. Default = wanted_taxon_seqs",
    required=False,
)


args = parser.parse_args()
frags_fp = args.input_frags
iccm = args.input_ccmetagen
tax_level = args.taxonomic_level
tax_name = args.taxon
o_file = args.output_fp + ".fas"


# developing and debugging:
# frags_fp = "test2.frag"
# iccm = "/test2_CCMetagen_nt_results.csv"
# tax_name = "Aspergillaceae"
# tax_level = "Family"
# o_file = "wanted_seq.fas"

## Get a list for all Closest_match where <tax_rank> == <tax_name>:
df = pd.read_csv(iccm, sep=",", index_col=0)

wanted_df = df[df[tax_level].str.match(tax_name).fillna(False)]


# Then extract sequences for that taxon.
ofile_good = open(o_file, "w")

cg = 0  # count good reads
with open(frags_fp) as frags:
    for line in csv.reader(frags, delimiter="\t"):
        match = line[5]
        seq_id = line[6]
        seq_read = line[0]

        if match in wanted_df.index.values:
            ofile_good.write(">" + seq_id + "\n" + seq_read + "\n")
            cg += 1

ofile_good.close()

print("")
print("Done.")
print(
    "%i sequences classified as %s were saved in the %s file." % (cg, tax_name, o_file)
)
print("Forward and reverse reads were saved to the same file")
print("")



#!/usr/bin/env ipython
# -*- coding: utf-8 -*-

"""
Functions to parse KMA results

@ V.R.Marcelino
Created on 1 Aug 2018.
"""

import re

# local imports
from ccmetagen import cTaxInfo, fNCBItax

# function to filter a res file in pandas df format:
def res_filter(df, cov, Iden, Depth, p):
    df = df.drop(df[df.Template_Coverage < cov].index)

    # filter based on identity
    df = df.drop(df[df.Query_Identity < Iden].index)

    # filter based on depth
    df = df.drop(df[df.Depth < Depth].index)

    # filter based on p-values
    df = df.drop(df[df.p_value > p].index)

    return df


# function that takes as input a pandas dataframe with KMA results
# and add tax information to results
def populate_w_tax(
    in_df,
    ref_database,
    species_threshold,
    genus_threshold,
    family_threshold,
    order_threshold,
    class_threshold,
    phylum_threshold,
    taxfile=None,
):
    # For default thresholds, see ccmetagen/__init__.py

    # Make sure all taxa columns are strings (doesn't automatically happen if the first one is None)
    in_df = in_df.assign(
        LCA_TaxId="",
        Superkingdom="",
        Kingdom="",
        Phylum="",
        Class="",
        Order="",
        Family="",
        Genus="",
        Species="",
    )

    # index == the #template (fungal match)
    for index, row in in_df.iterrows():
        match_info = cTaxInfo.TaxInfo()

        # define the tax. rank based on similarity:
        if ref_database == "UNITE":
            split_match = re.split(r"(\|| )", index)
            qiden = row["Query_Identity"]
            match_info.Lineage = split_match[12]

            # if taxid is knwon:
            if split_match[4] != "unk_taxid":
                match_info.TaxId = int(split_match[4])
                match_info = fNCBItax.lineage_extractor(
                    match_info.TaxId, match_info, taxfile
                )

                # Warning about unknown taxids:
            else:
                print("")
                print(
                    "WARNING: based on accession number, no taxonomic information was found in NCBI for %s"
                    % (match_info.Lineage)
                )
                print("This match will not get NCBItax taxonomic ranks")
                print("")
                match_info.TaxId = split_match[4]  # 'unk_taxid'

        elif ref_database == "RefSeq":
            split_match = re.split(r"(\|| )", index)
            qiden = row["Query_Identity"]
            match_info.TaxId = int(split_match[4])
            species = split_match[6] + " " + split_match[8]
            match_info.Lineage = species
            # include info from NCBI:
            match_info = fNCBItax.lineage_extractor(
                match_info.TaxId, match_info, taxfile
            )

        elif ref_database == "nt":
            split_match = re.split(r"(\|| )", index)
            qiden = row["Query_Identity"]
            match_info.Lineage = split_match[2]

            # get taxid from accession number
            taxid = split_match[0]

            if taxid == "unk_taxid":
                # Warning about unknown taxids:
                print("")
                print(
                    "WARNING: no NCBI's taxid found for accession %s"
                    % (match_info.Lineage)
                )
                print("This match will not get taxonomic ranks")
                print("")

            else:
                match_info.TaxId = int(taxid)
                match_info = fNCBItax.lineage_extractor(
                    match_info.TaxId, match_info, taxfile
                )

        # Populate the df with lineage info and the LCA taxid:
        in_df.at[index, "Superkingdom"] = match_info.Superkingdom
        in_df.at[index, "Kingdom"] = match_info.Kingdom

        # Assign LCA_taxid. Go to Kingdom if possible:
        in_df.at[index, "LCA_TaxId"] = match_info.Superkingdom_TaxId

        if match_info.Kingdom_TaxId is not None:
            in_df.at[index, "LCA_TaxId"] = match_info.Kingdom_TaxId

        # if it matches to uncultured or unclassified fungus, use the Fungi LCA itaxid:
        if match_info.Kingdom == "Fungi":
            in_df.at[index, "LCA_TaxId"] = 4751

        thresholds = {
            "Phylum": phylum_threshold,
            "Class": class_threshold,
            "Order": order_threshold,
            "Family": family_threshold,
            "Genus": genus_threshold,
            "Species": species_threshold
        }

        for rank, threshold in thresholds.items():
            if qiden >= threshold:
                tax_info_attr = f"{rank}_TaxId"

                in_df.at[index, rank] = getattr(match_info, rank)

                if getattr(match_info, tax_info_attr) is not None:
                    in_df.at[index, "LCA_TaxId"] = getattr(match_info, tax_info_attr)

    return in_df
